# Copyright Allen Institute for Artificial Intelligence 2017
"""
ai2thor.server

Handles all communication with Unity through a Flask service.  Messages
are sent to the controller using a pair of request/response queues.
"""


import json
import logging
import sys
import os
import os.path

try:
    from queue import Empty
except ImportError:
    from Queue import Empty

import time
import warnings

from flask import Flask, request, make_response, abort
import werkzeug
import werkzeug.serving
import werkzeug.http
import numpy as np
from enum import Enum

from ai2thor.util.depth import apply_real_noise, generate_noise_indices

logging.getLogger('werkzeug').setLevel(logging.ERROR)

werkzeug.serving.WSGIRequestHandler.protocol_version = 'HTTP/1.1'

MAX_DEPTH = 5000

# get with timeout to allow quit
def queue_get(que, unity_proc=None):
    res = None
    while True:
        try:
            res = que.get(block=True, timeout=0.5)
            break
        except Empty:
            # we poll here for the unity proc in the event that it has
            # exited otherwise we would wait indefinetly for the queue
            if unity_proc and unity_proc.poll() is not None:
                raise Exception("Unity process exited %s" % unity_proc.returncode)
            pass
    return res

class NumpyAwareEncoder(json.JSONEncoder):

    def default(self, obj):
        if isinstance(obj, np.generic):
            return np.asscalar(obj)
        return super(NumpyAwareEncoder, self).default(obj)

class BufferedIO(object):
    def __init__(self, wfile):
        self.wfile = wfile
        self.data = []

    def write(self, output):
        self.data.append(output)

    def flush(self):
        self.wfile.write(b"".join(self.data))
        self.wfile.flush()
    def close(self):
        return self.wfile.close()

    @property
    def closed(self):
        return self.wfile.closed

class ThorRequestHandler(werkzeug.serving.WSGIRequestHandler):
    def run_wsgi(self):
        old_wfile = self.wfile
        self.wfile = BufferedIO(self.wfile)
        result = super(ThorRequestHandler, self).run_wsgi()
        self.wfile = old_wfile
        return result

class MultiAgentEvent(object):

    def __init__(self, active_agent_id, events):
        self._active_event = events[active_agent_id]
        self.metadata = self._active_event.metadata
        self.screen_width = self._active_event.screen_width
        self.screen_height = self._active_event.screen_height
        self.events = events
        self.third_party_camera_frames = []
        # XXX add methods for depth,sem_seg

    @property
    def cv2img(self):
        return self._active_event.cv2img

    def add_third_party_camera_image(self, third_party_image_data):
        self.third_party_camera_frames.append(read_buffer_image(third_party_image_data, self.screen_width, self.screen_height))


def read_buffer_image(buf, width, height, flip_y=True, flip_x=False, dtype=np.uint8,
                      flip_rb_colors=False):
    im_bytes = np.frombuffer(buf.tobytes(), dtype=dtype) if sys.version_info.major < 3 \
        else np.frombuffer(buf, dtype=dtype)
    im = im_bytes.reshape(height, width, -1)
    if flip_y:
        im = np.flip(im, axis=0)
    if flip_x:
        im = np.flip(im, axis=1)
    if flip_rb_colors:
        im = im[..., ::-1]

    return im


def unique_rows(arr, return_index=False, return_inverse=False):
    arr = np.ascontiguousarray(arr).copy()
    b = arr.view(np.dtype((np.void, arr.dtype.itemsize * arr.shape[1])))
    if return_inverse:
        _, idx, inv = np.unique(b, return_index=True, return_inverse=True)
    else:
        _, idx = np.unique(b, return_index=True)
    unique = arr[idx]
    if return_index and return_inverse:
        return unique, idx, inv
    elif return_index:
        return unique, idx
    elif return_inverse:
        return unique, inv
    else:
        return unique

class Event(object):
    """
    Object that is returned from a call to  controller.step().
    This class wraps the screenshot that Unity captures as well
    as the metadata sent about each object
    """

    def __init__(self, metadata):
        self.metadata = metadata
        self.screen_width = metadata['screenWidth']
        self.screen_height = metadata['screenHeight']

        self.frame = None
        self.depth_frame = None
        self.normals_frame = None
        self.flow_frame = None

        self.color_to_object_id = {}
        self.object_id_to_color = {}

        self.instance_detections2D = None
        self.instance_masks = {}
        self.class_masks = {}

        self.instance_segmentation_frame = None
        self.class_segmentation_frame = None

        self.class_detections2D = {}

        self.process_colors()
        self.process_visible_bounds2D()
        self.third_party_camera_frames = []
        self.third_party_class_segmentation_frames = []
        self.third_party_instance_segmentation_frames = []
        self.third_party_depth_frames = []
        self.third_party_normals_frames = []
        self.third_party_flows_frames = []

        self.events = [self] # Ensure we have a similar API to MultiAgentEvent

    @property
    def image_data(self):
        warnings.warn("Event.image_data has been removed - RGB data can be retrieved from event.frame and encoded to an image format")
        return None

    def process_visible_bounds2D(self):
        if self.instance_detections2D and len(self.instance_detections2D) > 0:
            for obj in self.metadata['objects']:
                obj['visibleBounds2D'] = (obj['visible'] and obj['objectId'] in self.instance_detections2D)

    def process_colors(self):
        if 'colors' in self.metadata and self.metadata['colors']:
            for color_data in self.metadata['colors']:
                name = color_data['name']
                c_key = tuple(color_data['color'])
                self.color_to_object_id[c_key] = name
                self.object_id_to_color[name] = c_key

    def objects_by_type(self, object_type):
        return [obj for obj in self.metadata['objects'] if obj['objectType'] == object_type]

    def process_colors_ids(self):
        if self.instance_segmentation_frame is None:
            return

        MIN_DETECTION_LEN = 0

        self.instance_detections2D = {}
        unique_ids, unique_inverse = unique_rows(self.instance_segmentation_frame.reshape(-1, 3), return_inverse=True)
        unique_inverse = unique_inverse.reshape(self.instance_segmentation_frame.shape[:2])
        unique_masks = (np.tile(unique_inverse[np.newaxis, :, :], (len(unique_ids), 1, 1)) == np.arange(len(unique_ids))[:, np.newaxis, np.newaxis])
        #for unique_color_ind, unique_color in enumerate(unique_ids):
        for color_bounds in self.metadata['colorBounds']:
            color = np.array(color_bounds['color'])
            color_name = self.color_to_object_id.get(tuple(int(cc) for cc in color), 'background')
            cls = color_name
            simObj = False
            if '|' in cls:
                cls = cls.split('|')[0]
                simObj = True

            bb = np.array(color_bounds['bounds'])
            bb[[1,3]] = self.metadata['screenHeight'] - bb[[3,1]]
            if not((bb[2] - bb[0]) < MIN_DETECTION_LEN or (bb[3] - bb[1]) < MIN_DETECTION_LEN):
                if cls not in self.class_detections2D:
                    self.class_detections2D[cls] = []

                self.class_detections2D[cls].append(bb)

                color_ind = np.argmin(np.sum(np.abs(unique_ids - color), axis=1))

                if simObj:
                    self.instance_detections2D[color_name] = bb
                    self.instance_masks[color_name] = unique_masks[color_ind, ...]

                if cls not in self.class_masks:
                    self.class_masks[cls] = unique_masks[color_ind, ...]
                else:
                    self.class_masks[cls] = np.logical_or(self.class_masks[cls], unique_masks[color_ind, ...])

    def _image_depth(self, image_depth_data, **kwargs):
        image_depth = read_buffer_image(image_depth_data, self.screen_width, self.screen_height)
        depth_format = kwargs['depth_format']
        image_depth_out = image_depth[:,:,0] + image_depth[:,:,1] / np.float32(256) + image_depth[:,:,2] / np.float32(256 ** 2)
        multiplier = 1.0
        if depth_format != DepthFormat.Normalized:
            multiplier = kwargs['camera_far_plane'] - kwargs['camera_near_plane']
        elif depth_format == DepthFormat.Millimeters:
            multiplier *= 1000
        image_depth_out *= multiplier / 256.0

        depth_image_float = image_depth_out.astype(np.float32)

        if 'add_noise' in kwargs and kwargs['add_noise']:
            depth_image_float = apply_real_noise(
                depth_image_float,
                self.screen_width,
                indices=kwargs['noise_indices']
            )

        return depth_image_float

    def add_image_depth_robot(self, image_depth_data, depth_format, **kwargs):
        multiplier = 1.0
        camera_far_plane = kwargs.pop('camera_far_plane', 1)
        camera_near_plane = kwargs.pop('camera_near_plane', 0)
        if depth_format == DepthFormat.Normalized:
            multiplier = 1.0 / (camera_far_plane - camera_near_plane)
        elif depth_format == DepthFormat.Millimeters:
            multiplier = 1000.0

        image_depth = read_buffer_image(
            image_depth_data, self.screen_width, self.screen_height, **kwargs
        ).reshape(self.screen_height, self.screen_width) * multiplier
        self.depth_frame = image_depth.astype(np.float32)

    def add_image_depth(self, image_depth_data, **kwargs):
        self.depth_frame = self._image_depth(image_depth_data, **kwargs)

    def add_third_party_image_depth(self, image_depth_data, **kwargs):
        self.third_party_depth_frames.append(self._image_depth(image_depth_data, **kwargs))

    def add_third_party_image_normals(self, normals_data):
        self.third_party_normals_frames.append(read_buffer_image(normals_data, self.screen_width, self.screen_height))

    def add_image_normals(self, image_normals_data):
        self.normals_frame = read_buffer_image(image_normals_data, self.screen_width, self.screen_height)

    def add_third_party_image_flows(self, flows_data):
        self.third_party_flows_frames.append(read_buffer_image(flows_data, self.screen_width, self.screen_height))

    def add_image_flows(self, image_flows_data):
        self.flows_frame = read_buffer_image(image_flows_data, self.screen_width, self.screen_height)

    def add_third_party_camera_image(self, third_party_image_data):
        self.third_party_camera_frames.append(read_buffer_image(third_party_image_data, self.screen_width, self.screen_height))

    def add_image(self, image_data, **kwargs):
        self.frame = read_buffer_image(image_data, self.screen_width, self.screen_height, **kwargs)

    def add_image_ids(self, image_ids_data):
        self.instance_segmentation_frame = read_buffer_image(image_ids_data, self.screen_width, self.screen_height)
        self.process_colors_ids()

    def add_third_party_image_ids(self, image_ids_data):
        self.third_party_instance_segmentation_frames.append(read_buffer_image(image_ids_data, self.screen_width, self.screen_height))

    def add_image_classes(self, image_classes_data):
        self.class_segmentation_frame = read_buffer_image(image_classes_data, self.screen_width, self.screen_height)

    def add_third_party_image_classes(self, image_classes_data):
        self.third_party_class_segmentation_frames.append(read_buffer_image(image_classes_data, self.screen_width, self.screen_height))

    def cv2image(self):
        warnings.warn("Deprecated - please use event.cv2img")
        return self.cv2img

    @property
    def cv2img(self):
        return self.frame[...,::-1]

    @property
    def pose(self):
        agent_meta = self.metadata['agent']
        loc = agent_meta['position']
        rotation = round(agent_meta['rotation']['y'] * 1000)
        horizon = round(agent_meta['cameraHorizon'] * 1000)
        return (round(loc['x'] * 1000), round(loc['z'] * 1000), rotation, horizon)

    @property
    def pose_discrete(self):
        # XXX should have this as a parameter
        step_size = 0.25
        agent_meta = self.metadata['agent']
        loc = agent_meta['position']
        rotation = int(agent_meta['rotation']['y'] / 90.0)
        horizon = int(round(agent_meta['cameraHorizon']))
        return (int(loc['x'] / step_size), int(loc['z'] / step_size), rotation, horizon)

    def get_object(self, object_id):
        for obj in self.metadata['objects']:
            if obj['objectId'] == object_id:
                return obj
        return None


class MultipartFormParser(object):

    @staticmethod
    def get_boundary(request_headers):
        for h, value in request_headers:
            if h == 'Content-Type':
                ctype, ct_opts = werkzeug.http.parse_options_header(value)
                boundary = ct_opts['boundary'].encode('ascii')
                return boundary
        return None

    def __init__(self, data, boundary):

        self.form = {}
        self.files = {}

        full_boundary = b'--' + boundary 
        mid_boundary = b'\r\n' + full_boundary
        view = memoryview(data)
        i = data.find(full_boundary) + len(full_boundary)
        while i >= 0:
            next_offset = data.find(mid_boundary, i)
            if next_offset < 0:
                break
            headers_offset = i + 2 # add 2 for CRLF
            body_offset = data.find(b'\r\n\r\n', headers_offset)
            raw_headers = view[headers_offset: body_offset]
            body = view[body_offset + 4: next_offset]
            i = next_offset + len(mid_boundary)

            headers = {}
            for header in raw_headers.tobytes().decode('ascii').strip().split("\r\n"):

                k,v = header.split(':')
                headers[k.strip()] = v.strip()

            ctype, ct_opts = werkzeug.http.parse_options_header(headers['Content-Type'])
            cdisp, cd_opts = werkzeug.http.parse_options_header(headers['Content-disposition'])
            assert cdisp == 'form-data'

            if 'filename' in cd_opts:
                if cd_opts['name'] not in self.files:
                    self.files[cd_opts['name']] = []

                self.files[cd_opts['name']].append(body)

            else:
                if ctype == 'text/plain' and 'charset' in ct_opts:
                    body = body.tobytes().decode(ct_opts['charset'])
                if cd_opts['name'] not in self.form:
                    self.form[cd_opts['name']] = []

                self.form[cd_opts['name']].append(body)


class DepthFormat(Enum):
    Meters = 0,
    Normalized = 1,
    Millimeters = 2

class Server(object):

    def __init__(
            self,
            request_queue,
            response_queue,
            host,
            port=0,
            threaded=False,
            depth_format=DepthFormat.Meters,
            add_depth_noise=False,
            width=300,
            height=300
    ):

        app = Flask(__name__,
                    template_folder=os.path.realpath(
                        os.path.join(
                            os.path.dirname(os.path.abspath(__file__)), '..', 'templates')))

        self.image_buffer = None

        self.app = app
        self.client_token = None
        self.subscriptions = []
        self.app.config.update(PROPAGATE_EXCEPTIONS=False, JSONIFY_PRETTYPRINT_REGULAR=False)
        self.port = port
        self.last_rate_timestamp = time.time()
        self.frame_counter = 0
        self.debug_frames_per_interval = 50
        self.xwindow_id = None
        self.wsgi_server = werkzeug.serving.make_server(host, self.port, self.app, threaded=threaded, request_handler=ThorRequestHandler)
        # used to ensure that we are receiving frames for the action we sent
        self.sequence_id = 0
        self.last_event = None
        self.camera_near_plane = 0.1
        self.camera_far_plane = 20.0
        self.depth_format = depth_format
        self.add_depth_noise = add_depth_noise
        self.noise_indices = None

        if add_depth_noise:
            assert width == height,\
                "Noise supported with square dimension images only."
            self.noise_indices = generate_noise_indices(width)

        @app.route('/ping', methods=['get'])
        def ping():
            return 'pong'

        @app.route('/train', methods=['post'])
        def train():

            if request.headers['Content-Type'].split(';')[0] == 'multipart/form-data':
                form = MultipartFormParser(request.get_data(), MultipartFormParser.get_boundary(request.headers))
                metadata = json.loads(form.form['metadata'][0])
                token = form.form['token'][0]
            else:
                form = request
                metadata = json.loads(form.form['metadata'])
                token = form.form['token']

            if self.client_token and token != self.client_token:
                abort(403)

            if self.frame_counter % self.debug_frames_per_interval == 0:
                now = time.time()
                # rate = self.debug_frames_per_interval / float(now - self.last_rate_timestamp)
                self.last_rate_timestamp = now
                # import datetime
                # print("%s %s/s" % (datetime.datetime.now().isoformat(), rate))

            if metadata['sequenceId'] != self.sequence_id:
                raise ValueError("Sequence id mismatch: %s vs %s" % (
                    metadata['sequenceId'], self.sequence_id))

            events = []

            for i, a in enumerate(metadata['agents']):
                e = Event(a)
                image_mapping = dict(
                    image=e.add_image,
                    image_depth=lambda x: e.add_image_depth(
                        x,
                        depth_format=self.depth_format,
                        camera_near_plane=self.camera_near_plane,
                        camera_far_plane=self.camera_far_plane,
                        add_noise=self.add_depth_noise,
                        noise_indices=self.noise_indices
                    ),
                    image_ids=e.add_image_ids,
                    image_classes=e.add_image_classes,
                    image_normals=e.add_image_normals,
                    image_flows=e.add_image_flows
                )

                for key in image_mapping.keys():
                    if key in form.files:
                        image_mapping[key](form.files[key][i])

                third_party_image_mapping = dict(
                    image=e.add_image,
                    image_thirdParty_depth=lambda x: e.add_third_party_image_depth(
                        x,
                        depth_format=self.depth_format,
                        camera_near_plane=self.camera_near_plane,
                        camera_far_plane=self.camera_far_plane
                    ),
                    image_thirdParty_image_ids=e.add_third_party_image_ids,
                    image_thirdParty_classes=e.add_third_party_image_classes,
                    image_thirdParty_normals=e.add_third_party_image_normals,
                    image_thirdParty_flows=e.add_third_party_image_flows
                )

                if a['thirdPartyCameras'] is not None:
                    for ti, t in enumerate(a['thirdPartyCameras']):
                        for key in third_party_image_mapping.keys():
                            if key in form.files:
                                third_party_image_mapping[key](form.files[key][ti])
                events.append(e)

            if len(events) > 1:
                self.last_event = event = MultiAgentEvent(metadata['activeAgentId'], events)
            else:
                self.last_event = event = events[0]

            for img in form.files.get('image-thirdParty-camera', []):
                self.last_event.add_third_party_camera_image(img)

            request_queue.put_nowait(event)

            self.frame_counter += 1

            next_action = queue_get(response_queue)
            if 'sequenceId' not in next_action:
                self.sequence_id += 1
                next_action['sequenceId'] = self.sequence_id
            else:
                self.sequence_id = next_action['sequenceId']

            resp = make_response(json.dumps(next_action, cls=NumpyAwareEncoder))

            return resp

    def start(self):
        self.wsgi_server.serve_forever()

    def set_init_params(self, init_params):
        self.camera_near_plane = init_params['cameraNearPlane']
        self.camera_far_plane = init_params['cameraFarPlane']



from pprint import pprint
import json
import math
import copy


def vector_distance(v0, v1):
    dx = v0['x'] - v1['x']
    dy = v0['y'] - v1['y']
    dz = v0['z'] - v1['z']
    return math.sqrt(dx * dx + dy * dy + dz * dz)

def path_distance(path):
    distance = 0
    for i in range(0, len(path) - 1):
        distance += vector_distance(path[i], path[i+1])
    return distance


def compute_spl(episodes_with_golden):
    """
    Computes batch SPL from episode list
    :param episodes_with_golden:
        sequence of episode object, an episode should include the following keys:
                                 'path'
                                    the path to evaluate and 'shortest_path' the shortest path
                                    as returned by 'get_shortest_path_to_object'.
                                    Both as sequences with values of the form
                                    dict(x=float, y=float, z=float)
                                 'success' boolean, a 0 for a failed path 1 for a successful one
    :return: returns a float representing the spl
    """
    N = len(episodes_with_golden)
    eval_sum = 0.0
    for i, episode in enumerate(episodes_with_golden):
        path = episode['path']
        shortest_path = episode['shortest_path']
        eval_sum += compute_single_spl(path, shortest_path, episode['success'])
    return eval_sum / N

def compute_single_spl(path, shortest_path, successful_path):
    """
    Computes SPL for a path dict(x=float, y=float, z=float)
    :param path: Sequence of dict(x=float, y=float, z=float) representing the path to evaluate
    :param shortest_path: Sequence of dict(x=float, y=float, z=float) representing the shortest oath
    :param successful_path: boolean indicating if the path was successful, 0 for a failed path or 1 for a successful one
    :return:
    """
    Si = 1 if successful_path == True or successful_path == 1 else 0
    li = path_distance(shortest_path)
    pi = path_distance(path)
    spl = Si * (li / (max(pi, li)))
    return spl


def get_shortest_path_to_object(
        controller,
        object_id,
        initial_position,
        initial_rotation=None
    ):
    """
    Computes the shortest path to an object from an initial position using a controller
    :param controller: agent controller
    :param object_id: string with id of the object
    :param initial_position: dict(x=float, y=float, z=float) with the desired initial rotation
    :param initial_rotation: dict(x=float, y=float, z=float) representing rotation around axes or None
    :return:
    """
    args = dict(
            action='GetShortestPath',
            objectId=object_id,
            position=initial_position,
        )
    if initial_rotation is not None:
        args['rotation'] = initial_rotation
    event = controller.step(args)
    if event.metadata['lastActionSuccess']:
        return event.metadata['actionReturn']['corners']
    else:
        raise ValueError(
            "Unable to find shortest path for objectId '{}'".format(
                object_id
            )
        )


def get_shortest_path_to_object_type(
        controller,
        object_type,
        initial_position,
        initial_rotation=None
):
    """
    Computes the shortest path to an object from an initial position using a controller
    :param controller: agent controller
    :param object_type: string that represents the type of the object
    :param initial_position: dict(x=float, y=float, z=float) with the desired initial rotation
    :param initial_rotation: dict(x=float, y=float, z=float) representing rotation around axes or None
    """
    args = dict(
        action='GetShortestPath',
        objectType=object_type,
        position=initial_position
    )
    if initial_rotation is not None:
        args['rotation'] = initial_rotation
    event = controller.step(args)
    if event.metadata['lastActionSuccess']:
        return event.metadata['actionReturn']['corners']
    else:
        print(event.metadata['errorMessage'])
        raise ValueError(
            "Unable to find shortest path for object type '{}'".format(
                object_type
            )
        )


def get_shortest_path_to_point(
        controller,
        initial_position,
        target_position
    ):
    """
    Computes the shortest path to a point from an initial position using an agent controller
    :param controller: agent controller
    :param initial_position: dict(x=float, y=float, z=float) with the desired initial rotation
    :param target_position: dict(x=float, y=float, z=float) representing target position
    :return:
    """
    event = controller.step(
            action='GetShortestPathToPoint',
            position=initial_position,
            x=target_position['x'],
            y=target_position['y'],
            z=target_position['z']
        )
    if event.metadata['lastActionSuccess']:
        return event.metadata['actionReturn']['corners']
    else:
        raise ValueError(
            "Unable to find shortest path to point '{}'".format(
                target_position
            )
        )



def get_episodes_with_shortest_paths(controller, episodes, initialize_func=None):
    """
    Computes shortest path for an episode sequence
    :param controller: agent controller
    :param episodes: sequence of episode object required fields:
                        'target_object_id' string representing the object to look for
                        'initial_position' dict(x=float, y=float, z=float) of starting position
                        'initial_rotation' dict(x=float, y=float, z=float) representing rotation
                                           around axes
    :param initialize_func:
    :return:
    """
    episodes_with_golden = copy.deepcopy(episodes)
    for i, episode in enumerate(episodes_with_golden):
        event = controller.reset(episode['scene'])
        if initialize_func is not None:
            initialize_func()

        try:
            if 'target_object_id' in episode:
                episode['shortest_path'] = get_shortest_path_to_object(
                    controller,
                    episode['target_object_id'],
                    {
                        'x': episode['initial_position']['x'],
                        'y': episode['initial_position']['y'],
                        'z': episode['initial_position']['z']
                    },
                    {
                        'x': episode['initial_rotation']['x'],
                        'y': episode['initial_rotation']['y'],
                        'z': episode['initial_rotation']['z']
                    }
                )
            else:
                episode['shortest_path'] = get_shortest_path_to_object_type(
                    controller,
                    episode['target_object_type'],
                    {
                        'x': episode['initial_position']['x'],
                        'y': episode['initial_position']['y'],
                        'z': episode['initial_position']['z']
                    },
                    {
                        'x': episode['initial_rotation']['x'],
                        'y': episode['initial_rotation']['y'],
                        'z': episode['initial_rotation']['z']
                    }
                )
        except ValueError:
            raise ValueError(
                "Unable to find shortest path for objectId '{}' in episode '{}'".format(
                    episode['target_object_id'], json.dumps(episode, sort_keys=True, indent=4)
                )
            )
    return episodes_with_golden



import subprocess
import os.path
import re
import tempfile
from ai2thor.build import COMMIT_ID
import shlex
import shutil

def pci_records():
    records = []
    command = shlex.split('lspci -vmm')
    output = subprocess.check_output(command).decode()

    for devices in output.strip().split("\n\n"):
        record = {}
        records.append(record)
        for row in devices.split("\n"):
            key, value = row.split("\t")
            record[key.split(':')[0]] = value

    return records

def xorg_bus_id():
    bus_id = None
    for r in pci_records():
        if r.get('Vendor', '') == 'NVIDIA Corporation'\
                and r['Class'] in ['VGA compatible controller', '3D controller']:
            bus_id = 'PCI:' + ':'.join(map(lambda x: str(int(x, 16)), re.split(r'[:\.]', r['Slot'])))
            break

    if bus_id is None:
        raise Exception("no valid nvidia device could be found")

    return bus_id


def has_docker():
    with open(os.devnull, "w") as dn:
        return subprocess.call(['which', 'docker'], stdout=dn) == 0


def bridge_gateway():
    output = subprocess.check_output("docker network inspect -f '{{range .IPAM.Config}}{{.Gateway}}{{end}}' bridge", shell=True).decode('ascii').strip()
    if re.match(r'^[0-9]+\.[0-9]+\.[0-9]+\.[0-9]+$', output):
        return output
    else:
        raise Exception("Didn't receive a single ip address from network inspect bridge: %s" % output)


def nvidia_version():
    version = None
    version_path = '/proc/driver/nvidia/version'
    if os.path.isfile(version_path):
        with open(version_path) as f:
            for line in f:
                if line.startswith('NVRM version: NVIDIA'):
                    match = re.search(r'Kernel Module\s+([0-9\.]+)\s+', line)
                    if match:
                        version = match.group(1)
                        break
    return version


def generate_dockerfile(tag):

    driver_url = 'http://us.download.nvidia.com/XFree86/Linux-x86_64/{version}/NVIDIA-Linux-x86_64-{version}.run'.format(version=nvidia_version())
    driver_filename = os.path.basename(driver_url)

    dockerfile = """
FROM ai2thor/ai2thor-base:{tag}
RUN wget {driver_url} -P /root/
RUN sh /root/{driver_filename} -s --no-kernel-module
""".format(driver_filename=driver_filename, driver_url=driver_url, tag=tag)

    return dockerfile


def image_exists(image_name):
    output = subprocess.check_output("docker images -q %s" % image_name, shell=True)
    return len(output) > 0


def run(image_name, base_dir, command, environment):
    allowed_keys = {'AI2THOR_PORT', 'AI2THOR_CLIENT_TOKEN', 'AI2THOR_HOST'}

    environment_string = ""
    for k,v in environment.items():
        if k in allowed_keys:
            environment_string += " -e %s=%s " % (k,v)

    relative_command = os.path.relpath(command, base_dir)
    docker_command = os.path.join('/root/.ai2thor', relative_command)
    environment_string += " -e %s=%s -e %s='%s'" % ("AI2THOR_DEVICE_BUSID", xorg_bus_id(), "AI2THOR_COMMAND", docker_command)
    command = "docker run -v {base_dir}:/root/.ai2thor -d --privileged {environment} {image_name} /root/start.sh".format(
        environment=environment_string, image_name=image_name, base_dir=base_dir)
    container_id = subprocess.check_output(command, shell=True).decode('ascii').strip()
    return container_id


def kill_container(container_id):
    subprocess.check_output("docker kill %s" % container_id, shell=True)


def build_image():

    version = nvidia_version()
    image_name = 'ai2thor/ai2thor-nvidia-%s:%s' % (version, COMMIT_ID)

    if image_exists(image_name):
        return image_name

    td = tempfile.mkdtemp()
    with open("%s/Dockerfile" % td, "w") as f:
        f.write(generate_dockerfile(tag))

    subprocess.check_call("docker build --rm -t %s %s" % (image_name, td), shell=True)

    shutil.rmtree(td)

    return image_name



import copy

import glob
import shutil
import json
import pickle
import numpy as np
import cv2
from ai2thor.controller import key_for_point
import ai2thor.controller
from ai2thor.server import Event
from collections import defaultdict
import os

MOVE_AHEAD_MAP = {
    0: dict(x=0, z=1),
    90: dict(x=1, z=0),
    180: dict(x=0, z=-1),
    270: dict(x=-1, z=0),
    }

class Controller(object):

    def __init__(self, base_dir, grid_size=0.25):

        self.grid_size = grid_size
        self.base_dir = base_dir
        if grid_size < 0.25:
            raise Exception("must adjust granularity of key_for_point for smaller grid sizes")


    def reset(self, scene_name):
        self.scene_name = scene_name
        with open('%s/%s/index.json' % (self.base_dir,self.scene_name)) as f:
            self.positions = json.loads(f.read())

        #for g in glob.glob('%s/%s/metadata/*.json' % (self.base_dir,self.scene_name)):
        #    with open(g) as f:
        #        j = json.loads(f.read())
        #        pos = j['agent']['position']
        #        key = key_for_point(pos['x'], pos['z'])
        #        pos_id = os.path.splitext(os.path.basename(g))[0]
        #        event_path = os.path.join('%s/%s/events/%s.pickle' % (self.base_dir, self.scene_name, pos_id))
        #        self.positions[key].append({'event': event_path, 'metadata': j})

        #import random
        #total = len(self.positions)
        
        #p = self.positions[list(self.positions.keys())[random.randint(0, total - 1)]][3]
        #self.last_event = self.load_event(p)

    @property
    def position_x(self):
        return self.last_event.metadata['agent']['position']['x']

    @property
    def position_z(self):
        return self.last_event.metadata['agent']['position']['z']

    @property
    def rotation(self):
        return self.last_event.metadata['agent']['rotation']['y']

    @property
    def camera_horizon(self):
        return self.last_event.metadata['agent']['cameraHorizon']

    def start(self):
        pass
    
    def load_event(self, pos):
        with open(pos['event'], 'rb') as f:
            e = pickle.load(f)
        return e

    def find_position(self, x, z, rotation, camera_horizon):
        for p in self.positions.get(key_for_point(x, z), []):
            if abs(p['rotation'] - rotation) < 1.0 and abs(p['cameraHorizon'] - camera_horizon) < 1.0:
                event = self.load_event(p)
                return event

        return None

    def move(self, x, z):
        return self.find_position(x, z, self.rotation, self.camera_horizon)

    def move_back(self):
        m = MOVE_AHEAD_MAP[self.rotation]
        new_x = (-m['x'] * self.grid_size) + self.position_x
        new_z = (-m['z'] * self.grid_size) + self.position_z
        return self.move(new_x, new_z)

    def move_ahead(self):
        m = MOVE_AHEAD_MAP[self.rotation]
        new_x = (m['x'] * self.grid_size) + self.position_x
        new_z = (m['z'] * self.grid_size) + self.position_z
        return self.move(new_x, new_z)
    
    def look(self, new_horizon):
        if new_horizon < -30 or new_horizon > 30:
            return None
        else:
            return self.find_position(self.position_x, self.position_z, self.rotation, new_horizon)
            


    def look_up(self):
        return self.look(self.camera_horizon - 30)

    def look_down(self):
        return self.look(self.camera_horizon + 30)
    
    def rotate_right(self):
        new_rotation = (self.rotation + 90) % 360
        return self.rotate(new_rotation)

    def rotate(self, new_rotation):
        return self.find_position(self.position_x, self.position_z, new_rotation, self.camera_horizon)

    def rotate_left(self):
        new_rotation = (self.rotation - 90) % 360
        return self.rotate(new_rotation)

    def step(self, action=None, **action_args):
        if type(action) is dict:
            action = copy.deepcopy(action) # prevent changes from leaking
        else:
            action = dict(action=action)
       
        actions = dict(
            RotateRight=self.rotate_right,
            RotateLeft=self.rotate_left,
            MoveAhead=self.move_ahead,
            MoveBack=self.move_back,
            LookUp=self.look_up,
            LookDown=self.look_down
        )
         
        event = actions[action['action']]()
        if event is None:
            event = copy.deepcopy(self.last_event)
            event.metadata['lastActionSuccess'] = False
            event.metadata['lastAction'] = action['action']

        self.last_event = event
        return event

class FrameCounter:

    def __init__(self):
        self.counter = 0

    def inc(self):
        self.counter += 1

def write_frame(event, base_dir, scene_name, frame_name):
    import cv2
    import os
    import json
    events_dir = "%s/%s/events" % (base_dir, scene_name)
    met_dir = "%s/%s/metadata" % (base_dir, scene_name)
    os.makedirs(met_dir, exist_ok=True)
    os.makedirs(events_dir, exist_ok=True)

    with open(events_dir + "/%03d.pickle" % frame_name, "wb") as f:
        pickle.dump(event, f)

    with open(met_dir + "/%03d.json" % frame_name, "w") as f:
        f.write(json.dumps(event.metadata))

def look_up_down_write(controller, base_dir, fc, scene_name):

    fc.inc()
    write_frame(controller.step(action='LookUp'), base_dir, scene_name, fc.counter)
    fc.inc()
    write_frame(controller.step(action='LookDown'), base_dir, scene_name, fc.counter)
    fc.inc()
    write_frame(controller.step(action='LookDown'), base_dir, scene_name, fc.counter)
    controller.step(action='LookUp')

def index_metadata(base_dir, scene_name):
    positions_index = defaultdict(list)
    for g in glob.glob('%s/%s/metadata/*.json' % (base_dir, scene_name)):
        with open(g) as f:
            j = json.loads(f.read())
            agent = j['agent']
            pos = agent['position']
            key = key_for_point(pos['x'], pos['z'])
            pos_id = os.path.splitext(os.path.basename(g))[0]
            event_path = os.path.join('%s/%s/events/%s.pickle' % (base_dir, scene_name, pos_id))
            positions_index[key].append(
                    {'event': event_path, 'rotation': agent['rotation']['y'], 'cameraHorizon': agent['cameraHorizon']})
    with open("%s/%s/index.json" % (base_dir, scene_name), "w") as f:
        f.write(json.dumps(positions_index))


def dump_scene_controller(base_dir, controller):
    if controller.last_event is None:
        raise Exception("Controller must be reset and intialized to a scene")
    
    scene_name = controller.last_event.metadata['sceneName']
    fc = FrameCounter()

    shutil.rmtree("%s/%s" % (base_dir, scene_name), ignore_errors=True)

    event = controller.step(action='GetReachablePositions')
    for p in event.metadata['reachablePositions']:
        action = copy.deepcopy(p)
        action['action'] = 'TeleportFull'
        action['horizon'] = 0.0
        action['forceAction'] = True
        action['rotation'] = dict(y=0.0)
        event = controller.step(action)
        print(fc.counter)
        if event.metadata['lastActionSuccess']:
            look_up_down_write(controller, base_dir, fc, scene_name)
            for i in range(3):
                event = controller.step(action='RotateRight')
                look_up_down_write(controller, base_dir, fc, scene_name)
        

    index_metadata(base_dir, scene_name)


def dump_scene(scene_name, base_dir, renderObjectImage=False, renderDepthImage=False, renderClassImage=False):
    controller = ai2thor.controller.Controller()
    controller.start(height=448, width=448)
    controller.reset(scene_name) 
    event = controller.step(dict(action='Initialize', fieldOfView=90, gridSize=0.25, renderDepthImage=renderDepthImage, renderObjectImage=renderObjectImage, renderClassImage=renderClassImage))
    dump_scene_controller(base_dir, controller)
    controller.stop()




import numpy as np

REAL_2_SIM_TRANSFORM = np.array([
               [1.00854301, -0.0111386, 0.51920809],
               [0.00316833, 0.97336625, -1.15532594]
        ])


def transform_real_2_sim(real_position):
    """
    Transforms a position from the 'real' coordinate system to the 'sim' coordinate system.
    :param real_position: dictionary with 'x', 'y' and 'z' keys to floating point values
    :return: position in sim space as dictionary with 'x', 'y' and 'z' keys to floating point values
    """

    real_pos = np.array([real_position['x'], real_position['y'], 1])

    sim_pos_np = np.dot(
        REAL_2_SIM_TRANSFORM,
        real_pos
    )

    sim_pos = {
        'x': sim_pos_np[0],
        'y': 0.9010001,
        'z': sim_pos_np[1]
    }

    return sim_pos


import os
import time
import msgpack
import numpy as np
import requests
import cv2
from pprint import pprint
import shutil
import copy

import ai2thor.docker
import ai2thor.server
from ai2thor.server import queue_get
from ai2thor._quality_settings import QUALITY_SETTINGS, DEFAULT_QUALITY
from ai2thor.server import Event, MultiAgentEvent, DepthFormat
from ai2thor.interact import InteractiveControllerPrompt, DefaultActions


class Controller(object):

    def __init__(
            self,
            headless=False,
            host='127.0.0.1',
            port=0,
            width=300,
            height=300,
            agent_id=0,
            image_dir='.',
            save_image_per_frame=False,
            depth_format=DepthFormat.Meters,
            camera_near_plane=0.1,
            camera_far_plane=20
    ):
        self.host = host
        self.port = port
        self.headless = headless
        self.last_event = {}
        self.last_action = {}
        self.sequence_id = 0
        self.agent_id = agent_id
        self.screen_width = width
        self.screen_height = height
        self.depth_format = depth_format
        self.camera_near_plane = camera_near_plane,
        self.camera_far_plane = camera_far_plane

        if image_dir != '.':
            if os.path.exists(image_dir):
                shutil.rmtree(image_dir)
            os.makedirs(image_dir)

        self.interactive_controller = InteractiveControllerPrompt(
            [
                DefaultActions.MoveAhead,
                DefaultActions.MoveBack,
                DefaultActions.RotateLeft,
                DefaultActions.RotateRight,
                DefaultActions.LookUp,
                DefaultActions.LookDown
            ],
            image_dir=image_dir,
            image_per_frame=save_image_per_frame
        )

        self.start(
            port,
            host,
            agent_id=agent_id
        )

    def stop(self):
        pass

    def start(
            self,
            port=9200,
            host='127.0.0.1',
            agent_id=0,
            **kwargs
    ):
        self.host = host
        self.port = port
        self.agent_id = agent_id

        # response_payload = self._post_event('start')
        pprint('-- Start:')
        # pprint(response_payload)

    def reset(self, scene_name=None):
        if not scene_name.endswith('_physics'):
            scene_name = scene_name + "_physics"
        self.sequence_id = 0
        # response_payload = self._post_event(
        #     'reset', dict(action='Reset', sceneName=scene_name, sequenceId=self.sequence_id)
        # )
        pprint('-- Reset:')
        # pprint(response_payload)

        return self.last_event

    def step(self, action=None, **action_args):

        if type(action) is dict:
            action = copy.deepcopy(action) # prevent changes from leaking
        else:
            action = dict(action=action)

        raise_for_failure = action_args.pop('raise_for_failure', False)

        action.update(action_args)

        if self.headless:
            action["renderImage"] = False

        action['sequenceId'] = self.sequence_id
        action['agentId'] = self.agent_id

        self.last_action = action

        rotation = action.get('rotation')
        if rotation is not None and type(rotation) != dict:
            action['rotation'] = {}
            action['rotation']['y'] = rotation

        payload = self._post_event('step', action)
        events = []
        for i, agent_metadata in enumerate(payload['metadata']['agents']):
            event = Event(agent_metadata)
            image_mapping = dict(
                image=lambda x: event.add_image(x, flip_y=False, flip_rb_colors=False),
                image_depth=lambda x: event.add_image_depth_robot(
                    x,
                    self.depth_format,
                    camera_near_plane=self.camera_near_plane,
                    camera_far_plane=self.camera_far_plane,
                    flip_y=False,
                    dtype=np.float64
                )
            )
            for key in image_mapping.keys():
                if key in payload and len(payload[key]) > i:
                    image_mapping[key](payload[key][i])
            events.append(event)

        if len(events) > 1:
            self.last_event = MultiAgentEvent(self.agent_id, events)
        else:
            self.last_event = events[0]

        if not self.last_event.metadata['lastActionSuccess'] and self.last_event.metadata['errorCode'] == 'InvalidAction':
            raise ValueError(self.last_event.metadata['errorMessage'])

        if raise_for_failure:
            assert self.last_event.metadata['lastActionSuccess']

        # pprint("Display event:")
        # Controller._display_step_event(self.last_event)

        return self.last_event

    def interact(self,
                 class_segmentation_frame=False,
                 instance_segmentation_frame=False,
                 depth_frame=False,
                 color_frame=False,
                 metadata=False
                 ):
        self.interactive_controller.interact(
            self,
            class_segmentation_frame=class_segmentation_frame,
            instance_segmentation_frame=instance_segmentation_frame,
            depth_frame=depth_frame,
            color_frame=color_frame,
            metadata=metadata
        )

    def _post_event(self, route='', data=None):
        r = requests.post(self._get_url(route), json=data)
        s = time.time()
        pprint('ACTION "{}"'.format(data['action']))
        pprint("POST")
        # pprint(r.content)
        pprint(r.status_code)
        return msgpack.unpackb(r.content, raw=False)

    def _get_url(self, route=''):
        return 'http://{}:{}{}'.format(
            self.host,
            self.port,
            '/{}'.format(route) if route != '' else ''
        )

    @staticmethod
    def _display_step_event(event):
        metadata = event.metadata
        image = event.frame
        pprint(metadata)
        cv2.imshow('aoeu', event.cv2img)
        cv2.waitKey(1000)



from aws_requests_auth.boto_utils import BotoAWSRequestsAuth
import os
import requests

PUBLIC_S3_BUCKET = "ai2-thor-public"
PRIVATE_S3_BUCKET = "ai2-thor-private"

COMMIT_ID = None
try:
    from ai2thor._builds import COMMIT_ID
except ImportError:
    pass

platform_map = dict(Linux64="Linux", OSXIntel64="Darwin")

arch_platform_map = {v: k for k, v in platform_map.items()}


def boto_auth():
    return BotoAWSRequestsAuth(aws_host='s3-us-west-2.amazonaws.com',
            aws_region='us-west-2',
            aws_service='s3')

def build_name(arch, commit_id, include_private_scenes=False):
    if include_private_scenes:
        return "thor-private-%s-%s" % (arch, commit_id)
    else:
        return "thor-%s-%s" % (arch, commit_id)

base_url = "http://s3-us-west-2.amazonaws.com/%s/" % PUBLIC_S3_BUCKET
private_base_url = "http://s3-us-west-2.amazonaws.com/%s/" % PRIVATE_S3_BUCKET

class Build(object):

    def __init__(self, arch, commit_id, include_private_scenes):

        self.arch = arch
        self.commit_id = commit_id
        self.include_private_scenes = include_private_scenes

    def auth(self):
        if self.include_private_scenes:
            return boto_auth()
        else:
            return None

    def _base_url(self):
        if self.include_private_scenes:
            return private_base_url
        else:
            return base_url

    def url(self):
        return self._base_url() + os.path.join('builds', build_name(self.arch, self.commit_id, self.include_private_scenes) + ".zip")

    def log_url(self):
        return os.path.splitext(self.url())[0] + '.log'

    def sha256_url(self):
        return os.path.splitext(self.url())[0] + '.sha256'

    def exists(self):
        return requests.head(self.url(), auth=self.auth()).status_code == 200

    def log_exists(self):
        return requests.head(self.log_url(), auth=self.auth()).status_code == 200

    def sha256(self):
        res = requests.get(self.sha256_url(), auth=self.auth())
        res.raise_for_status()
        return res.content.decode('ascii')




import yaml
from pprint import pprint
from unityparser import UnityDocument

# Requires unity parser, run:
# pip install unityparser

def updateNavMeshParamsForScene(scene_file_name):
    pprint("Updating file '{}'...".format(scene_file_name))
    doc = UnityDocument.load_yaml(scene_file_name)
    for entry in doc.entries:
        if entry.__class__.__name__ == 'NavMeshSettings':
            # print(entry.__class__.__name__)
            buildSettings = getattr(entry, 'm_BuildSettings', None)
            # pprint(buildSettings)
            buildSettings['agentRadius'] = '0.175'
            buildSettings['agentHeight'] = '0.9'
            buildSettings['agentClimb'] = '0.5'
            buildSettings['manualCellSize'] = '1'

            buildSettings['cellSize'] = '0.03'

    doc.dump_yaml()


def GetSceneNames(last_index, last_subIndex, nameTemplate, prefix_path='unity/Assets/Scenes'):
    return ["{}/FloorPlan_{}{}_{}.unity".format(prefix_path, nameTemplate, i, j)  for i in range(1, last_index+1) for j in range(1, last_subIndex+1)]


def main():
    testSceneNames = GetSceneNames(3, 5, "Val")
    valSceneNames = GetSceneNames(2, 2, "test-dev", 'unity/Assets/Private/Scenes')
    trainSceneNames = GetSceneNames(12, 5, "Train")

    #allScenes = testSceneNames  + trainSceneNames
    allScenes = valSceneNames
    for scene_file_name in allScenes:
        updateNavMeshParamsForScene(scene_file_name)


if __name__== "__main__":
    main()


    # Exceptions:
    # Scene FloorPlan_Train7_1
    # Train_11_3 unmade bed
    # Val2_3 unamde bed


import tty
import sys
import termios
from PIL import Image
import numpy as np
import os

from enum import Enum
import json


class DefaultActions(Enum):
    MoveRight = 0,
    MoveLeft = 1,
    MoveAhead = 2,
    MoveBack = 3,
    LookUp = 4,
    LookDown = 5,
    RotateRight = 8,
    RotateLeft = 9

# TODO tie this with actions
# class ObjectActions(Enum):
#     PutObject
#     MoveHandAhead
#     MoveHandBack
#     MoveHandRight
#     MoveHandLeft
#     MoveHandUp
#     MoveHandDown
#     DropHandObject
#     PickupObject,
#     OpenObject,
#     CloseObject,
#     ToggleObjectOff


def get_term_character():
    fd = sys.stdin.fileno()
    old_settings = termios.tcgetattr(fd)
    try:
        tty.setraw(sys.stdin.fileno())
        ch = sys.stdin.read(1)
    finally:
        termios.tcsetattr(fd, termios.TCSADRAIN, old_settings)
    return ch


class InteractiveControllerPrompt(object):
    def __init__(self, default_actions, has_object_actions=True, image_dir='.', image_per_frame=False):
        self.default_actions = default_actions
        self.has_object_actions = has_object_actions
        self.image_per_frame = image_per_frame
        self.image_dir = image_dir
        self.counter = 0

        default_interact_commands = {
            '\x1b[C': dict(action='MoveRight', moveMagnitude=0.25),
            '\x1b[D': dict(action='MoveLeft', moveMagnitude=0.25),
            '\x1b[A': dict(action='MoveAhead', moveMagnitude=0.25),
            '\x1b[B': dict(action='MoveBack', moveMagnitude=0.25),
            '\x1b[1;2A': dict(action='LookUp'),
            '\x1b[1;2B': dict(action='LookDown'),
            'i': dict(action='LookUp'),
            'k': dict(action='LookDown'),
            'l': dict(action='RotateRight'),
            'j': dict(action='RotateLeft'),
            '\x1b[1;2C': dict(action='RotateRight'),
            '\x1b[1;2D': dict(action='RotateLeft')
        }
        action_set = {a.name for a in default_actions}

        self.default_interact_commands = {
            k: v for (k, v) in default_interact_commands.items() if v['action'] in action_set
        }

    def interact(self,
                 controller,
                 class_segmentation_frame=False,
                 instance_segmentation_frame=False,
                 depth_frame=False,
                 color_frame=False,
                 metadata=False
                 ):

        if not sys.stdout.isatty():
            raise RuntimeError("controller.interact() must be run from a terminal")

        default_interact_commands = self.default_interact_commands

        self._interact_commands = default_interact_commands.copy()

        command_message = u"Enter a Command: Move \u2190\u2191\u2192\u2193, Rotate/Look Shift + \u2190\u2191\u2192\u2193, Quit 'q' or Ctrl-C"
        print(command_message)
        for a in self.next_interact_command():
            new_commands = {}
            command_counter = dict(counter=1)

            def add_command(cc, action, **args):
                if cc['counter'] < 15:
                    com = dict(action=action)
                    com.update(args)
                    new_commands[str(cc['counter'])] = com
                    cc['counter'] += 1

            event = controller.step(a)
            visible_objects = []
            InteractiveControllerPrompt.write_image(
                event,
                self.image_dir,
                "_{}".format(self.counter),
                image_per_frame=self.image_per_frame,
                class_segmentation_frame=class_segmentation_frame,
                instance_segmentation_frame=instance_segmentation_frame,
                color_frame=color_frame,
                depth_frame=depth_frame,
                metadata=metadata
            )

            self.counter += 1
            if self.has_object_actions:
                for o in event.metadata['objects']:
                    if o['visible']:
                        visible_objects.append(o['objectId'])
                        if o['openable']:
                            if o['isOpen']:
                                add_command(command_counter, 'CloseObject', objectId=o['objectId'])
                            else:
                                add_command(command_counter, 'OpenObject', objectId=o['objectId'])

                        if o['toggleable']:
                            add_command(command_counter, 'ToggleObjectOff', objectId=o['objectId'])

                        if len(event.metadata['inventoryObjects']) > 0:
                            inventoryObjectId = event.metadata['inventoryObjects'][0]['objectId']
                            if o['receptacle'] and (not o['openable'] or o['isOpen']) and inventoryObjectId != o['objectId']:
                                add_command(command_counter, 'PutObject', objectId=inventoryObjectId, receptacleObjectId=o['objectId'])
                                add_command(command_counter, 'MoveHandAhead', moveMagnitude=0.1)
                                add_command(command_counter, 'MoveHandBack', moveMagnitude=0.1)
                                add_command(command_counter, 'MoveHandRight', moveMagnitude=0.1)
                                add_command(command_counter, 'MoveHandLeft', moveMagnitude=0.1)
                                add_command(command_counter, 'MoveHandUp', moveMagnitude=0.1)
                                add_command(command_counter, 'MoveHandDown', moveMagnitude=0.1)
                                add_command(command_counter, 'DropHandObject')

                        elif o['pickupable']:
                            add_command(command_counter, 'PickupObject', objectId=o['objectId'])

            self._interact_commands = default_interact_commands.copy()
            self._interact_commands.update(new_commands)

            print("Position: {}".format(event.metadata['agent']['position']))
            print(command_message)
            print("Visible Objects:\n" + "\n".join(sorted(visible_objects)))

            skip_keys = ['action', 'objectId']
            for k in sorted(new_commands.keys()):
                v = new_commands[k]
                command_info = [k + ")", v['action']]
                if 'objectId' in v:
                    command_info.append(v['objectId'])

                for a, av in v.items():
                    if a in skip_keys:
                        continue
                    command_info.append("%s: %s" % (a, av))

                print(' '.join(command_info))

    def next_interact_command(self):
        current_buffer = ''
        while True:
            commands = self._interact_commands
            current_buffer += get_term_character()
            if current_buffer == 'q' or current_buffer == '\x03':
                break

            if current_buffer in commands:
                yield commands[current_buffer]
                current_buffer = ''
            else:
                match = False
                for k,v in commands.items():
                    if k.startswith(current_buffer):
                        match = True
                        break

                if not match:
                    current_buffer = ''

    @classmethod
    def write_image(
            cls,
            event,
            image_dir,
            suffix,
            image_per_frame=False,
            class_segmentation_frame=False,
            instance_segmentation_frame=False,
            depth_frame=False,
            color_frame=False,
            metadata=False
    ):

        def save_image(name, image, flip_br=False):
            # TODO try to use PIL which did not work with RGBA
            # image.save(
            #     name
            # )
            import cv2
            img = image
            if flip_br:
                img = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
            cv2.imwrite("{}.png".format(name), img)

        def array_to_image(arr, mode=None):
            return arr

        def json_write(name, obj):
            with open("{}.json".format(name), 'w') as outfile:
                json.dump(obj, outfile, indent=4, sort_keys=True)

        frame_writes = [
            ('color',
             color_frame,
             lambda event: event.frame,
             array_to_image,
             lambda x, y: save_image(x, y, flip_br=True)
             ),
            ('instance_segmentation',
             instance_segmentation_frame,
             lambda event: event.instance_segmentation_frame,
             array_to_image,
             save_image
             ),
            ('class_segmentation',
             class_segmentation_frame,
             lambda event: event.class_segmentation_frame,
             array_to_image,
             save_image
             ),
            ('depth',
             depth_frame,
             lambda event: event.depth_frame,
             lambda data: array_to_image(
                 (255.0 / data.max() * (data - data.min())).astype(np.uint8)
             ),
             save_image
             ),
            ('depth_raw',
             depth_frame,
             lambda event: event.depth_frame,
             lambda x: x,
             lambda name, x: np.save(
                 name.strip('.png').strip('./') if image_dir is '.' else name.strip('.png'),
                 x.astype(np.float32)
             )
             ),
            ('metadata',
             metadata,
             lambda event: event.metadata,
             lambda x: x,
             json_write
             )
        ]

        for frame_filename, condition, frame_func, transform, save in frame_writes:
            frame = frame_func(event)
            if frame is not None:
                frame = transform(frame)
                image_name = os.path.join(
                    image_dir,
                    "{}{}"
                        .format(
                        frame_filename,
                        "{}".format(suffix) if image_per_frame else ""
                    )
                )
                print("Image {}, {}".format(image_name, image_dir))
                save(image_name, frame)

            else:
                print("No frame present, call initialize with the right parameters")



import os
import numpy as np

dir_path = os.path.dirname(os.path.realpath(__file__))

def generate_noise_indices(img_size):
    img_size = int(img_size)
    default_size = 300
    corners = np.load(os.path.join(dir_path, "depth_noise.npy"), allow_pickle=True)
    indices = []
    for j, corner in enumerate(corners):
        height_indices = np.array([], dtype=np.int32)
        width_indices = np.array([], dtype=np.int32)

        if img_size != default_size:
            idx = 0 if j <= 1 else len(corner) - 1
            width = corner[idx]
            height = len(corner)

            w_ratio = width / default_size
            h_ratio = height / default_size

            width = int(round(w_ratio * img_size))
            height = int(round(h_ratio * img_size))

            m = (height - 0) / (0 - width)
            b = height
            t = np.array([], dtype=np.int32)
            for y in range(height):
                x = (y - b) / m
                t = np.append(t, int(round(x)))

            t = np.flip(t, 0) if j > 1 else t
            corner = t

        for i, c in enumerate(corner):

            offset = 0
            i_offset = 0
            if j % 2 != 0:
                offset = img_size - c
            if j > 1:
                i_offset = img_size - len(corner)
            x = np.repeat(i_offset + i, c)
            height_indices = np.concatenate((height_indices, x))
            y = np.array(range(offset, offset + c))
            width_indices = np.concatenate((width_indices, y))

            indices.append((
                    np.array(height_indices, dtype=np.int32),
                    np.array(width_indices, dtype=np.int32)
                )
            )

    return indices


def apply_real_noise(depth_arr, size, indices=None):
    """
    Applies noise to depth image to look more similar to the real depth camera
    :param depth_arr: numpy square 2D array representing the depth
    :param size: square size of array
    :param indices: cached indices where noise is going to be applied, if None they get calculated
                    here based on the image size.
    :return:
    """
    if indices is None:
        indices = generate_noise_indices(size)
    for index_arr in indices:
        depth_arr[index_arr] = 0.0
    return depth_arr



from flask import Flask, request, jsonify, make_response
import numpy as np
import io
import threading
import random
import msgpack
import werkzeug
import werkzeug.serving
import werkzeug.http
import os


def random_image(height, width):
    img = np.zeros(height * width * 3, dtype=np.uint8).reshape(height, width, 3)
    img[:,:,0] = random.randint(0, 255)
    img[:,:,1] = random.randint(0, 255)
    img[:,:,2] = random.randint(0, 255)

    return img


class MockServer(object):

    def __init__(self, height, width):
        app = Flask(__name__)

        self.height = height
        self.width = width


        @app.route('/ping', methods=['GET'])
        def ping():
            return make_response('PONG')

        @app.route('/start', methods=['POST'])
        def start():
            return make_response(msgpack.packb({'status': 200}, use_bin_type=True))

        @app.route('/reset', methods=['POST'])
        def reset():
            return make_response(msgpack.packb({'status': 200}, use_bin_type=True))

        @app.route('/step', methods=['POST'])
        def step():
            content = request.json

            metadata = {
                u'sequenceId': content['sequenceId'],
                u'agents': [{
                    u'agentId': 0,
                    u'screenHeight': self.height,
                    u'screenWidth': self.width,
                    u'lastAction': content['action'],
                    u'lastActionSuccess': True
                }]
            }

            result = {
                u'image': [random_image(self.height, self.width).tostring()],
                u'image_depth': [],
                u'metadata': metadata}
            out = msgpack.packb(result, use_bin_type=True)

            return make_response(out)

        self.host = '127.0.0.1'
        self.port = 9200
        app.config.update(PROPAGATE_EXCEPTIONS=True, JSONIFY_PRETTYPRINT_REGULAR=False)
        self.wsgi_server = werkzeug.serving.make_server(self.host, self.port, app)
    
    def start(self):
        self.wsgi_server.serve_forever()





# Copyright Allen Institute for Artificial Intelligence 2017
"""
ai2thor.controller

Primary entrypoint into the Thor API. Provides all the high-level functions
needed to control the in-game agent through ai2thor.server.

"""
import atexit
from collections import deque, defaultdict
from itertools import product
import io
import json
import copy
import logging
import fcntl
import math
import time
import random
import shlex
import signal
import subprocess
import shutil
import re
import threading
import os
import platform
import uuid
import tty
import sys
import termios
import fcntl
try:
    from queue import Queue
except ImportError:
    from Queue import Queue

import zipfile

import numpy as np

import ai2thor.docker
import ai2thor.downloader
import ai2thor.server
from ai2thor.interact import InteractiveControllerPrompt, DefaultActions
from ai2thor.server import queue_get, DepthFormat
from ai2thor.build import COMMIT_ID
from ai2thor._quality_settings import QUALITY_SETTINGS, DEFAULT_QUALITY

import warnings

logger = logging.getLogger(__name__)

RECEPTACLE_OBJECTS = {
    'Box': {'Candle',
            'CellPhone',
            'Cloth',
            'CreditCard',
            'Dirt',
            'KeyChain',
            'Newspaper',
            'ScrubBrush',
            'SoapBar',
            'SoapBottle',
            'ToiletPaper'},
    'Cabinet': {'Bowl',
                'BowlDirty',
                'Box',
                'Bread',
                'BreadSliced',
                'ButterKnife',
                'Candle',
                'CellPhone',
                'Cloth',
                'CoffeeMachine',
                'Container',
                'ContainerFull',
                'CreditCard',
                'Cup',
                'Fork',
                'KeyChain',
                'Knife',
                'Laptop',
                'Mug',
                'Newspaper',
                'Pan',
                'Plate',
                'Plunger',
                'Pot',
                'Potato',
                'Sandwich',
                'ScrubBrush',
                'SoapBar',
                'SoapBottle',
                'Spoon',
                'SprayBottle',
                'Statue',
                'TissueBox',
                'Toaster',
                'ToiletPaper',
                'WateringCan'},
    'CoffeeMachine': {'MugFilled', 'Mug'},
    'CounterTop': {'Apple',
                   'AppleSlice',
                   'Bowl',
                   'BowlDirty',
                   'BowlFilled',
                   'Box',
                   'Bread',
                   'BreadSliced',
                   'ButterKnife',
                   'Candle',
                   'CellPhone',
                   'CoffeeMachine',
                   'Container',
                   'ContainerFull',
                   'CreditCard',
                   'Cup',
                   'Egg',
                   'EggFried',
                   'EggShell',
                   'Fork',
                   'HousePlant',
                   'KeyChain',
                   'Knife',
                   'Laptop',
                   'Lettuce',
                   'LettuceSliced',
                   'Microwave',
                   'Mug',
                   'MugFilled',
                   'Newspaper',
                   'Omelette',
                   'Pan',
                   'Plate',
                   'Plunger',
                   'Pot',
                   'Potato',
                   'PotatoSliced',
                   'RemoteControl',
                   'Sandwich',
                   'ScrubBrush',
                   'SoapBar',
                   'SoapBottle',
                   'Spoon',
                   'SprayBottle',
                   'Statue',
                   'Television',
                   'TissueBox',
                   'Toaster',
                   'ToiletPaper',
                   'Tomato',
                   'TomatoSliced',
                   'WateringCan'},
    'Fridge': {'Apple',
               'AppleSlice',
               'Bowl',
               'BowlDirty',
               'BowlFilled',
               'Bread',
               'BreadSliced',
               'Container',
               'ContainerFull',
               'Cup',
               'Egg',
               'EggFried',
               'EggShell',
               'Lettuce',
               'LettuceSliced',
               'Mug',
               'MugFilled',
               'Omelette',
               'Pan',
               'Plate',
               'Pot',
               'Potato',
               'PotatoSliced',
               'Sandwich',
               'Tomato',
               'TomatoSliced'},
    'GarbageCan': {'Apple',
                   'AppleSlice',
                   'Box',
                   'Bread',
                   'BreadSliced',
                   'Candle',
                   'CellPhone',
                   'CreditCard',
                   'Egg',
                   'EggFried',
                   'EggShell',
                   'LettuceSliced',
                   'Newspaper',
                   'Omelette',
                   'Plunger',
                   'Potato',
                   'PotatoSliced',
                   'Sandwich',
                   'ScrubBrush',
                   'SoapBar',
                   'SoapBottle',
                   'SprayBottle',
                   'Statue',
                   'ToiletPaper',
                   'Tomato',
                   'TomatoSliced'},
    'Microwave': {'Bowl',
                  'BowlDirty',
                  'BowlFilled',
                  'Bread',
                  'BreadSliced',
                  'Container',
                  'ContainerFull',
                  'Cup',
                  'Egg',
                  'EggFried',
                  'Mug',
                  'MugFilled',
                  'Omelette',
                  'Plate',
                  'Potato',
                  'PotatoSliced',
                  'Sandwich'},
    'PaintingHanger': {'Painting'},
    'Pan': {'Apple',
            'AppleSlice',
            'EggFried',
            'Lettuce',
            'LettuceSliced',
            'Omelette',
            'Potato',
            'PotatoSliced',
            'Tomato',
            'TomatoSliced'},
    'Pot': {'Apple',
            'AppleSlice',
            'EggFried',
            'Lettuce',
            'LettuceSliced',
            'Omelette',
            'Potato',
            'PotatoSliced',
            'Tomato',
            'TomatoSliced'},
    'Sink': {'Apple',
             'AppleSlice',
             'Bowl',
             'BowlDirty',
             'BowlFilled',
             'ButterKnife',
             'Container',
             'ContainerFull',
             'Cup',
             'Egg',
             'EggFried',
             'EggShell',
             'Fork',
             'Knife',
             'Lettuce',
             'LettuceSliced',
             'Mug',
             'MugFilled',
             'Omelette',
             'Pan',
             'Plate',
             'Pot',
             'Potato',
             'PotatoSliced',
             'Sandwich',
             'ScrubBrush',
             'SoapBottle',
             'Spoon',
             'Tomato',
             'TomatoSliced',
             'WateringCan'},
    'StoveBurner': {'Omelette', 'Pot', 'Pan', 'EggFried'},
    'TableTop': {'Apple',
                 'AppleSlice',
                 'Bowl',
                 'BowlDirty',
                 'BowlFilled',
                 'Box',
                 'Bread',
                 'BreadSliced',
                 'ButterKnife',
                 'Candle',
                 'CellPhone',
                 'CoffeeMachine',
                 'Container',
                 'ContainerFull',
                 'CreditCard',
                 'Cup',
                 'Egg',
                 'EggFried',
                 'EggShell',
                 'Fork',
                 'HousePlant',
                 'KeyChain',
                 'Knife',
                 'Laptop',
                 'Lettuce',
                 'LettuceSliced',
                 'Microwave',
                 'Mug',
                 'MugFilled',
                 'Newspaper',
                 'Omelette',
                 'Pan',
                 'Plate',
                 'Plunger',
                 'Pot',
                 'Potato',
                 'PotatoSliced',
                 'RemoteControl',
                 'Sandwich',
                 'ScrubBrush',
                 'SoapBar',
                 'SoapBottle',
                 'Spoon',
                 'SprayBottle',
                 'Statue',
                 'Television',
                 'TissueBox',
                 'Toaster',
                 'ToiletPaper',
                 'Tomato',
                 'TomatoSliced',
                 'WateringCan'},
    'ToiletPaperHanger': {'ToiletPaper'},
    'TowelHolder': {'Cloth'}}


def get_term_character():
    fd = sys.stdin.fileno()
    old_settings = termios.tcgetattr(fd)
    try:
        tty.setraw(sys.stdin.fileno())
        ch = sys.stdin.read(1)
    finally:
        termios.tcsetattr(fd, termios.TCSADRAIN, old_settings)
    return ch

def process_alive(pid):
    """
    Use kill(0) to determine if pid is alive
    :param pid: process id
    :rtype: bool
    """
    try:
        os.kill(pid, 0)
    except OSError:
        return False

    return True

# python2.7 compatible makedirs
def makedirs(directory):
    if not os.path.isdir(directory):
        os.makedirs(directory)

def distance(point1, point2):
    x_diff = (point1['x'] - point2['x']) ** 2
    z_diff = (point1['z'] - point2['z']) ** 2
    return math.sqrt(x_diff + z_diff)


def key_for_point(x, z):
    return "%0.1f %0.1f" % (x, z)

class Controller(object):

    def __init__(
            self,
            quality=DEFAULT_QUALITY,
            fullscreen=False,
            headless=False,
            port=0,
            start_unity=True,
            local_executable_path=None,
            width=300,
            height=300,
            x_display=None,
            host='127.0.0.1',
            scene='FloorPlan_Train1_1',
            image_dir='.',
            save_image_per_frame=False,
            docker_enabled=False,
            depth_format=DepthFormat.Meters,
            add_depth_noise=False,
            download_only=False,
            include_private_scenes=False,
            **unity_initialization_parameters
    ):
        self.request_queue = Queue(maxsize=1)
        self.response_queue = Queue(maxsize=1)
        self.receptacle_nearest_pivot_points = {}
        self.server = None
        self.unity_proc = None
        self.unity_pid = None
        self.docker_enabled = docker_enabled
        self.container_id = None
        self.local_executable_path = local_executable_path
        self.last_event = None
        self._scenes_in_build = None
        self.server_thread = None
        self.killing_unity = False
        self.quality = quality
        self.lock_file = None
        self.fullscreen = fullscreen
        self.headless = headless
        self.depth_format = depth_format
        self.add_depth_noise = add_depth_noise
        self.include_private_scenes = include_private_scenes


        self.interactive_controller = InteractiveControllerPrompt(
            list(DefaultActions),
            has_object_actions=True,
            image_dir=image_dir,
            image_per_frame=save_image_per_frame
        )

        if download_only:
            self.download_binary()
        else:
            self.start(
                port=port,
                start_unity=start_unity,
                width=width,
                height=height,
                x_display=x_display,
                host=host
            )

            self.initialization_parameters = unity_initialization_parameters

            if 'continuous' in self.initialization_parameters:
                warnings.warn(
                    "Warning: 'continuous' is deprecated and will be ignored,"
                    " use 'snapToGrid={}' instead."
                    .format(not self.initialization_parameters['continuous']),
                    DeprecationWarning
                )

            if 'continuousMode' in self.initialization_parameters:
                warnings.warn(
                    "Warning: 'continuousMode' is deprecated and will be ignored,"
                    " use 'snapToGrid={}' instead."
                        .format(not self.initialization_parameters['continuousMode']),
                    DeprecationWarning
                )

            event = self.reset(scene)
            if event.metadata['lastActionSuccess']:
                init_return = event.metadata['actionReturn']
                self.server.set_init_params(init_return)

                print("Initialize return: {}".format(init_return))
            else:
                raise RuntimeError('Initialize action failure: {}'.format(
                    event.metadata['errorMessage'])
                )

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.stop()

    @property
    def scenes_in_build(self):
        if self._scenes_in_build:
            return self._scenes_in_build

        event = self.step(action='GetScenesInBuild')

        self._scenes_in_build = set(event.metadata['actionReturn'])

        return self._scenes_in_build

    def reset(self, scene='FloorPlan_Train1_1'):
        if re.match(r'^FloorPlan[0-9]+$', scene):
            scene = scene + "_physics"

        if scene not in self.scenes_in_build:
            def key_sort_func(scene_name):
                m = re.search('FloorPlan[_]?([a-zA-Z\-]*)([0-9]+)_?([0-9]+)?.*$', scene_name)
                last_val = m.group(3) if m.group(3) is not None else -1
                return m.group(1), int(m.group(2)), int(last_val)
            raise ValueError(
                "\nScene '{}' not contained in build (scene names are case sensitive)."
                "\nPlease choose one of the following scene names:\n\n{}".format(
                    scene,
                    ", ".join(sorted(list(self.scenes_in_build), key=key_sort_func))
                )
            )

        self.response_queue.put_nowait(dict(action='Reset', sceneName=scene, sequenceId=0))
        self.last_event = queue_get(self.request_queue, self.unity_proc)  # can this be deleted?
        self.last_event = self.step(action='Initialize', **self.initialization_parameters)

        return self.last_event

    def random_initialize(
            self,
            random_seed=None,
            randomize_open=False,
            unique_object_types=False,
            exclude_receptacle_object_pairs=[],
            max_num_repeats=1,
            remove_prob=0.5):

        if random_seed is None:
            random_seed = random.randint(0, 2**32)

        exclude_object_ids = []

        for obj in self.last_event.metadata['objects']:
            pivot_points = self.receptacle_nearest_pivot_points
            # don't put things in pot or pan currently
            if (pivot_points and obj['receptacle'] and
                    pivot_points[obj['objectId']].keys()) or obj['objectType'] in ['Pot', 'Pan']:

                #print("no visible pivots for receptacle %s" % o['objectId'])
                exclude_object_ids.append(obj['objectId'])

        return self.step(dict(
            action='RandomInitialize',
            randomizeOpen=randomize_open,
            uniquePickupableObjectTypes=unique_object_types,
            excludeObjectIds=exclude_object_ids,
            excludeReceptacleObjectPairs=exclude_receptacle_object_pairs,
            maxNumRepeats=max_num_repeats,
            removeProb=remove_prob,
            randomSeed=random_seed))

    def scene_names(self):
        scenes = []
        for low, high in [(1,31), (201, 231), (301, 331), (401, 431)]:
            for i in range(low, high):
                scenes.append('FloorPlan%s_physics' % i)

        return scenes

    def unlock_release(self):
        if self.lock_file:
            fcntl.flock(self.lock_file, fcntl.LOCK_UN)

    def lock_release(self):
        build_dir = os.path.join(self.releases_dir(), self.build_name())
        if os.path.isdir(build_dir):
            self.lock_file = open(os.path.join(build_dir, ".lock"), "w")
            fcntl.flock(self.lock_file, fcntl.LOCK_SH)

    def prune_releases(self):
        current_exec_path = self.executable_path()
        for d in os.listdir(self.releases_dir()):
            release = os.path.join(self.releases_dir(), d)

            if current_exec_path.startswith(release):
                continue

            if os.path.isdir(release):
                try:
                    with open(os.path.join(release, ".lock"), "w") as f:
                        fcntl.flock(f, fcntl.LOCK_EX | fcntl.LOCK_NB)
                        shutil.rmtree(release)
                except Exception:
                    pass

    def next_interact_command(self):
        current_buffer = ''
        while True:
            commands = self._interact_commands
            current_buffer += get_term_character()
            if current_buffer == 'q' or current_buffer == '\x03':
                break

            if current_buffer in commands:
                yield commands[current_buffer]
                current_buffer = ''
            else:
                match = False
                for k,v in commands.items():
                    if k.startswith(current_buffer):
                        match = True
                        break

                if not match:
                    current_buffer = ''

    def interact(self,
                 class_segmentation_frame=False,
                 instance_segmentation_frame=False,
                 depth_frame=False,
                 color_frame=False,
                 metadata=False
                 ):
        self.interactive_controller.interact(
            self,
            class_segmentation_frame,
            instance_segmentation_frame,
            depth_frame,
            color_frame,
            metadata
        )

    def multi_step_physics(self, action, timeStep=0.05, max_steps=20):
        events = []
        self.step(action=dict(action='PausePhysicsAutoSim'), raise_for_failure=True)
        events.append(self.step(action))
        while not self.last_event.metadata['isSceneAtRest']:
            events.append(
                self.step(action=dict(
                    action='AdvancePhysicsStep',
                    timeStep=timeStep), raise_for_failure=True))

            if len(events) == (max_steps - 1):
                events.append(self.step(action=dict(action='UnpausePhysicsAutoSim'), raise_for_failure=True))
                break

        return events

    def step(self, action=None, **action_args):

        if type(action) is dict:
            action = copy.deepcopy(action) # prevent changes from leaking
        else:
            action = dict(action=action)

        raise_for_failure = action_args.pop('raise_for_failure', False)
        action.update(action_args)

        if self.headless:
            action["renderImage"] = False

        # prevent changes to the action from leaking
        action = copy.deepcopy(action)

        # XXX should be able to get rid of this with some sort of deprecation warning
        if 'AI2THOR_VISIBILITY_DISTANCE' in os.environ:
            action['visibilityDistance'] = float(os.environ['AI2THOR_VISIBILITY_DISTANCE'])

        should_fail = False
        self.last_action = action

        if ('objectId' in action and (action['action'] == 'OpenObject' or action['action'] == 'CloseObject')):

            force_visible = action.get('forceVisible', False)
            if not force_visible and self.last_event.instance_detections2D and action['objectId'] not in self.last_event.instance_detections2D:
                should_fail = True

            obj_metadata = self.last_event.get_object(action['objectId'])
            if obj_metadata is None or obj_metadata['isOpen'] == (action['action'] == 'OpenObject'):
                should_fail = True


        rotation = action.get('rotation')
        if rotation is not None and type(rotation) != dict:
            action['rotation'] = {}
            action['rotation']['y'] = rotation

        if should_fail:
            new_event = copy.deepcopy(self.last_event)
            new_event.metadata['lastActionSuccess'] = False
            self.last_event = new_event
            return new_event

        assert self.request_queue.empty()

        self.response_queue.put_nowait(action)
        self.last_event = queue_get(self.request_queue, self.unity_proc)

        if not self.last_event.metadata['lastActionSuccess'] and self.last_event.metadata['errorCode'] == 'InvalidAction':
            raise ValueError(self.last_event.metadata['errorMessage'])

        if raise_for_failure:
            assert self.last_event.metadata['lastActionSuccess']

        return self.last_event

    def unity_command(self, width, height, headless):
        command = self.executable_path()
        if headless:
            command += " -batchmode"
        else:
            fullscreen = 1 if self.fullscreen else 0
            if QUALITY_SETTINGS[self.quality] == 0:
                raise RuntimeError("Quality {} is associated with an index of 0. "
                                   "Due to a bug in unity, this quality setting would be ignored.".format(self.quality))
            command += " -screen-fullscreen %s -screen-quality %s -screen-width %s -screen-height %s" % (fullscreen, QUALITY_SETTINGS[self.quality], width, height)
        return shlex.split(command)

    def _start_unity_thread(self, env, width, height, host, port, image_name):
        # get environment variables

        env['AI2THOR_CLIENT_TOKEN'] = self.server.client_token = str(uuid.uuid4())
        env['AI2THOR_HOST'] = host
        env['AI2THOR_PORT'] = str(port)

        env['AI2THOR_SERVER_SIDE_SCREENSHOT'] = 'False' if self.headless else 'True'

        # print("Viewer: http://%s:%s/viewer" % (host, port))
        command = self.unity_command(width, height, headless=self.headless)

        if image_name is not None:
            self.container_id = ai2thor.docker.run(image_name, self.base_dir(), ' '.join(command), env)
            atexit.register(lambda: ai2thor.docker.kill_container(self.container_id))
        else:
            self.unity_proc = proc = subprocess.Popen(command, env=env)
            self.unity_pid = proc.pid
            atexit.register(lambda: proc.poll() is None and proc.kill())

    def check_docker(self):
        if self.docker_enabled:
            assert ai2thor.docker.has_docker(), "Docker enabled, but could not find docker binary in path"
            assert ai2thor.docker.nvidia_version() is not None,\
                "No nvidia driver version found at /proc/driver/nvidia/version - Dockerized THOR is only \
                    compatible with hosts with Nvidia cards with a driver installed"

    def check_x_display(self, x_display):
        with open(os.devnull, "w") as dn:
            # copying the environment so that we pickup
            # XAUTHORITY values
            env = os.environ.copy()
            env['DISPLAY'] = x_display

            if subprocess.call(['which', 'xdpyinfo'], stdout=dn) == 0:
                assert subprocess.call("xdpyinfo", stdout=dn, env=env, shell=True) == 0, \
                    ("Invalid DISPLAY %s - cannot find X server with xdpyinfo" % x_display)

    def _start_server_thread(self):
        self.server.start()

    def releases_dir(self):
        return os.path.join(self.base_dir(), 'releases')

    def base_dir(self):
        return os.path.join(os.path.expanduser('~'), '.ai2thor')

    def build_url(self):
        from ai2thor.build import arch_platform_map
        import ai2thor.build
        arch = arch_platform_map[platform.system()]
        if COMMIT_ID:
            ver_build = ai2thor.build.Build(arch, COMMIT_ID, self.include_private_scenes)
            return (ver_build.url(), ver_build.sha256())
        else:
            url = None
            sha256_build = None
            git_dir = os.path.normpath(os.path.dirname(os.path.realpath(__file__)) + "/../.git")
            for commit_id in subprocess.check_output('git --git-dir=' + git_dir + ' log -n 10 --format=%H', shell=True).decode('ascii').strip().split("\n"):
                commit_build = ai2thor.build.Build(arch, commit_id, self.include_private_scenes)

                try:
                    u = commit_build.url()
                    if os.path.isfile(self.executable_path(url=u)):
                        # don't need sha256 since we aren't going to download
                        url = u
                        break
                    elif commit_build.exists():
                        sha256_build = commit_build.sha256()
                        url = u
                        break
                except Exception:
                    pass

            if url is None:
                raise Exception("Couldn't find a suitable build url for platform: %s" % platform.system())

            # print("Got build for %s: " % (url))

            return (url, sha256_build)

    def build_name(self, url=None):
        if url is None:
            url, _ = self.build_url()
        return os.path.splitext(os.path.basename(url))[0]

    def executable_path(self, url=None):

        if self.local_executable_path is not None:
            return self.local_executable_path

        target_arch = platform.system()

        bn = self.build_name(url)
        if target_arch == 'Linux':
            return os.path.join(self.releases_dir(), bn, bn)
        elif target_arch == 'Darwin':
            return os.path.join(
                self.releases_dir(),
                bn,
                bn + ".app",
                "Contents/MacOS",
                "AI2-Thor")
        else:
            raise Exception('unable to handle target arch %s' % target_arch)

    def download_binary(self):

        if self.local_executable_path:
            return

        if platform.architecture()[0] != '64bit':
            raise Exception("Only 64bit currently supported")

        url, sha256_build = self.build_url()
        tmp_dir = os.path.join(self.base_dir(), 'tmp')
        makedirs(self.releases_dir())
        makedirs(tmp_dir)
        download_lf = open(os.path.join(tmp_dir, self.build_name(url) + ".lock"), "w")
        try:
            fcntl.flock(download_lf, fcntl.LOCK_EX)

            if not os.path.isfile(self.executable_path()):
                zip_data = ai2thor.downloader.download(
                    url,
                    self.build_name(),
                    sha256_build,
                    self.include_private_scenes
                    )

                z = zipfile.ZipFile(io.BytesIO(zip_data))
                # use tmpdir instead or a random number
                extract_dir = os.path.join(tmp_dir, self.build_name())
                logger.debug("Extracting zipfile %s" % os.path.basename(url))
                z.extractall(extract_dir)
                os.rename(extract_dir, os.path.join(self.releases_dir(), self.build_name()))
                # we can lose the executable permission when unzipping a build
                os.chmod(self.executable_path(), 0o755)
            else:
                logger.debug("%s exists - skipping download" % self.executable_path())

        finally:
            fcntl.flock(download_lf, fcntl.LOCK_UN)
            download_lf.close()

    def start(
            self,
            port=0,
            start_unity=True,
            width=300,
            height=300,
            x_display=None,
            host='127.0.0.1',
            player_screen_width=None,
            player_screen_height=None
    ):

        if 'AI2THOR_VISIBILITY_DISTANCE' in os.environ:
            import warnings
            warnings.warn("AI2THOR_VISIBILITY_DISTANCE environment variable is deprecated, use \
                the parameter visibilityDistance parameter with the Initialize action instead")

        if player_screen_width is not None:
            warnings.warn("'player_screen_width' parameter is deprecated, use the 'width'"
                          " parameter instead.")
            width = player_screen_width

        if player_screen_height is not None:
            warnings.warn("'player_screen_height' parameter is deprecated, use the 'height'"
                          " parameter instead.")
            height = player_screen_height

        if height <= 0 or width <= 0:
            raise Exception("Screen resolution must be > 0x0")

        if self.server_thread is not None:
            import warnings
            warnings.warn('start method depreciated. The server started when the Controller was initialized.')

            # Stops the current server and creates a new one. This is done so
            # that the arguments passed in will be used on the server.
            self.stop()

        env = os.environ.copy()

        image_name = None

        if self.docker_enabled:
            self.check_docker()
            host = ai2thor.docker.bridge_gateway()

        self.server = ai2thor.server.Server(
            self.request_queue,
            self.response_queue,
            host,
            port=port,
            depth_format=self.depth_format,
            add_depth_noise=self.add_depth_noise,
            width=width,
            height=height
        )


        self.server_thread = threading.Thread(target=self._start_server_thread)

        self.server_thread.daemon = True
        self.server_thread.start()

        if start_unity:
            if platform.system() == 'Linux':

                if self.docker_enabled:
                    image_name = ai2thor.docker.build_image()
                else:

                    if x_display:
                        env['DISPLAY'] = ':' + x_display
                    elif 'DISPLAY' not in env:
                        env['DISPLAY'] = ':0.0'

                    self.check_x_display(env['DISPLAY'])

            if not self.local_executable_path:
                self.download_binary()
                self.lock_release()
                self.prune_releases()

            _, port = self.server.wsgi_server.socket.getsockname()
            self._start_unity_thread(env, width, height, host, port, image_name)

        # receive the first request
        self.last_event = queue_get(self.request_queue, self.unity_proc)

        if height < 300 or width < 300:
            self.last_event = self.step('ChangeResolution', x=width, y=height)

        return self.last_event

    def stop(self):
        self.response_queue.put_nowait({})
        self.stop_unity()
        self.server.wsgi_server.shutdown()
        self.stop_container()
        self.unlock_release()

    def stop_container(self):
        if self.container_id:
            ai2thor.docker.kill_container(self.container_id)
            self.container_id = None

    def stop_unity(self):
        if self.unity_pid and process_alive(self.unity_pid):
            self.killing_unity = True
            os.kill(self.unity_pid, signal.SIGTERM)
            for i in range(10):
                if not process_alive(self.unity_pid):
                    break
                time.sleep(0.1)
            if process_alive(self.unity_pid):
                os.kill(self.unity_pid, signal.SIGKILL)

class BFSSearchPoint:
    def __init__(self, start_position, move_vector, heading_angle=0.0, horizon_angle=0.0):
        self.start_position = start_position
        self.move_vector = defaultdict(lambda: 0.0)
        self.move_vector.update(move_vector)
        self.heading_angle = heading_angle
        self.horizon_angle = horizon_angle

    def target_point(self):
        x = self.start_position['x'] + self.move_vector['x']
        z = self.start_position['z'] + self.move_vector['z']
        return dict(x=x, z=z)

class BFSController(Controller):

    def __init__(self, grid_size=0.25):
        super(BFSController, self).__init__()
        self.rotations = [0, 90, 180, 270]
        self.horizons = [330, 0, 30]
        self.allow_enqueue = True
        self.queue = deque()
        self.seen_points = []
        self.visited_seen_points = []
        self.grid_points = []
        self.grid_size = grid_size
        self._check_visited = False
        self.distance_threshold = self.grid_size / 5.0

    def visualize_points(self, scene_name, wait_key=10):
        import cv2
        points = set()
        xs = []
        zs = []

        # Follow the file as it grows
        for point in self.grid_points:
            xs.append(point['x'])
            zs.append(point['z'])
            points.add(str(point['x']) + "," + str(point['z']))

        image_width = 470
        image_height = 530
        image = np.zeros((image_height, image_width, 3), np.uint8)
        if not xs:
            return

        min_x = min(xs) - 1
        max_x = max(xs) + 1
        min_z = min(zs) - 1
        max_z = max(zs) + 1

        for point in list(points):
            x, z = map(float, point.split(','))
            circle_x = round(((x - min_x) / float(max_x - min_x)) * image_width)
            z = (max_z - z) + min_z
            circle_y = round(((z - min_z) / float(max_z - min_z)) * image_height)
            cv2.circle(image, (circle_x, circle_y), 5, (0, 255, 0), -1)

        cv2.imshow(scene_name, image)
        cv2.waitKey(wait_key)

    def has_islands(self):
        queue = []
        seen_points = set()
        mag = self.grid_size

        def enqueue_island_points(p):
            if json.dumps(p) in seen_points:
                return
            queue.append(dict(z=p['z'] + mag, x=p['x']))
            queue.append(dict(z=p['z'] - mag, x=p['x']))
            queue.append(dict(z=p['z'], x=p['x'] + mag))
            queue.append(dict(z=p['z'], x=p['x'] - mag))
            seen_points.add(json.dumps(p))

        enqueue_island_points(self.grid_points[0])

        while queue:
            point_to_find = queue.pop()
            for p in self.grid_points:
                dist = math.sqrt(
                    ((point_to_find['x'] - p['x']) ** 2) +
                    ((point_to_find['z'] - p['z']) ** 2))

                if dist < 0.05:
                    enqueue_island_points(p)

        return len(seen_points) != len(self.grid_points)

    def build_graph(self):
        import networkx as nx
        graph = nx.Graph()
        for point in self.grid_points:
            self._build_graph_point(graph, point)

        return graph

    def key_for_point(self, point):
        return "{x:0.3f}|{z:0.3f}".format(**point)

    def _build_graph_point(self, graph, point):
        for p in self.grid_points:
            dist = math.sqrt(((point['x'] - p['x']) ** 2) + ((point['z'] - p['z']) ** 2))
            if dist <= (self.grid_size + 0.01) and dist > 0:
                graph.add_edge(self.key_for_point(point), self.key_for_point(p))

    def move_relative_points(self, all_points, graph, position, rotation):

        action_orientation = {
            0:dict(x=0, z=1, action='MoveAhead'),
            90:dict(x=1, z=0, action='MoveRight'),
            180:dict(x=0, z=-1, action='MoveBack'),
            270:dict(x=-1, z=0, action='MoveLeft')
        }

        move_points = dict()

        for n in graph.neighbors(self.key_for_point(position)):
            point = all_points[n]
            x_o = round((point['x'] - position['x']) / self.grid_size)
            z_o = round((point['z'] - position['z']) / self.grid_size)
            for target_rotation, offsets in action_orientation.items():
                delta = round(rotation + target_rotation) % 360
                ao = action_orientation[delta]
                action_name = action_orientation[target_rotation]['action']
                if x_o == ao['x'] and z_o == ao['z']:
                    move_points[action_name] = point
                    break

        return move_points

    def plan_horizons(self, agent_horizon, target_horizon):
        actions = []
        horizon_step_map = {330:3, 0:2, 30:1, 60:0}
        look_diff = horizon_step_map[int(agent_horizon)] - horizon_step_map[int(target_horizon)]
        if look_diff > 0:
            for i in range(look_diff):
                actions.append(dict(action='LookDown'))
        else:
            for i in range(abs(look_diff)):
                actions.append(dict(action='LookUp'))

        return actions

    def plan_rotations(self, agent_rotation, target_rotation):
        right_diff = target_rotation - agent_rotation
        if right_diff < 0:
            right_diff += 360
        right_steps = right_diff / 90

        left_diff = agent_rotation - target_rotation
        if left_diff < 0:
            left_diff += 360
        left_steps = left_diff / 90

        actions = []
        if right_steps < left_steps:
            for i in range(int(right_steps)):
                actions.append(dict(action='RotateRight'))
        else:
            for i in range(int(left_steps)):
                actions.append(dict(action='RotateLeft'))

        return actions

    def shortest_plan(self, graph, agent, target):
        import networkx as nx
        path = nx.shortest_path(graph, self.key_for_point(agent['position']), self.key_for_point(target['position']))
        actions = []
        all_points = {}

        for point in self.grid_points:
            all_points[self.key_for_point(point)] = point

        #assert all_points[path[0]] == agent['position']

        current_position = agent['position']
        current_rotation = agent['rotation']['y']

        for p in path[1:]:
            inv_pms = {self.key_for_point(v): k for k, v in self.move_relative_points(all_points, graph, current_position, current_rotation).items()}
            actions.append(dict(action=inv_pms[p]))
            current_position = all_points[p]

        actions += self.plan_horizons(agent['cameraHorizon'], target['cameraHorizon'])
        actions += self.plan_rotations(agent['rotation']['y'], target['rotation']['y'])
        # self.visualize_points(path)

        return actions

    def enqueue_point(self, point):

        # ensure there are no points near the new point
        if self._check_visited or not any(map(lambda p: distance(p, point.target_point()) < self.distance_threshold, self.seen_points)):
            self.seen_points.append(point.target_point())
            self.queue.append(point)

    def enqueue_points(self, agent_position):

        if not self.allow_enqueue:
            return

        if not self._check_visited or not any(map(lambda p: distance(p, agent_position) < self.distance_threshold, self.visited_seen_points)):
            self.enqueue_point(BFSSearchPoint(agent_position, dict(x=-1 * self.grid_size)))
            self.enqueue_point(BFSSearchPoint(agent_position, dict(x=self.grid_size)))
            self.enqueue_point(BFSSearchPoint(agent_position, dict(z=-1 * self.grid_size)))
            self.enqueue_point(BFSSearchPoint(agent_position, dict(z=1 * self.grid_size)))
            self.visited_seen_points.append(agent_position)

    def search_all_closed(self, scene_name):
        self.allow_enqueue = True
        self.queue = deque()
        self.seen_points = []
        self.visited_seen_points = []
        self.grid_points = []
        event = self.reset(scene_name)
        event = self.step(dict(action='Initialize', gridSize=self.grid_size))
        self.enqueue_points(event.metadata['agent']['position'])
        while self.queue:
            self.queue_step()
            # self.visualize_points(scene_name)

    def start_search(
            self,
            scene_name,
            random_seed,
            full_grid,
            current_receptacle_object_pairs,
            randomize=True):

        self.seen_points = []
        self.visited_seen_points = []
        self.queue = deque()
        self.grid_points = []

        # we only search a pre-defined grid with all the cabinets/fridges closed
        # then keep the points that can still be reached
        self.allow_enqueue = True

        for gp in full_grid:
            self.enqueue_points(gp)

        self.allow_enqueue = False

        self.reset(scene_name)
        receptacle_object_pairs = []
        for op in current_receptacle_object_pairs:
            object_id, receptacle_object_id = op.split('||')
            receptacle_object_pairs.append(
                dict(receptacleObjectId=receptacle_object_id,
                     objectId=object_id))

        if randomize:
            self.random_initialize(
                random_seed=random_seed,
                unique_object_types=True,
                exclude_receptacle_object_pairs=receptacle_object_pairs)

        # there is some randomization in initialize scene
        # and if a seed is passed in this will keep it
        # deterministic
        if random_seed is not None:
            random.seed(random_seed)

        self.initialize_scene()
        while self.queue:
            self.queue_step()
            #self.visualize_points(scene_name)

        self.prune_points()
        #self.visualize_points(scene_name)

    # get rid of unreachable points
    def prune_points(self):
        final_grid_points = set()

        for gp in self.grid_points:
            final_grid_points.add(key_for_point(gp['x'], gp['z']))

        pruned_grid_points = []

        for gp in self.grid_points:
            found = False
            for x in [1, -1]:
                found |= key_for_point(gp['x'] + (self.grid_size * x), gp['z']) in final_grid_points

            for z in [1, -1]:
                found |= key_for_point(
                    gp['x'],
                    (self.grid_size * z) + gp['z']) in final_grid_points

            if found:
                pruned_grid_points.append(gp)

        self.grid_points = pruned_grid_points

    def is_object_visible(self, object_id):
        for obj in self.last_event.metadata['objects']:
            if obj['objectId'] == object_id and obj['visible']:
                return True
        return False

    def find_visible_receptacles(self):
        receptacle_points = []
        receptacle_pivot_points = []

        # pickup all objects
        visibility_object_id = None
        visibility_object_types = ['Mug', 'CellPhone']
        for obj in self.last_event.metadata['objects']:
            if obj['pickupable']:
                self.step(action=dict(
                    action='PickupObject',
                    objectId=obj['objectId'],
                    forceVisible=True))
            if visibility_object_id is None and obj['objectType'] in visibility_object_types:
                visibility_object_id = obj['objectId']

        for point in self.grid_points:
            self.step(dict(
                action='Teleport',
                x=point['x'],
                y=point['y'],
                z=point['z']), raise_for_failure=True)

            for rot, hor in product(self.rotations, self.horizons):
                event = self.step(
                    dict(action='RotateLook', rotation=rot, horizon=hor),
                    raise_for_failure=True)
                for j in event.metadata['objects']:
                    if j['receptacle'] and j['visible']:
                        receptacle_points.append(dict(
                            distance=j['distance'],
                            pivotId=0,
                            receptacleObjectId=j['objectId'],
                            searchNode=dict(
                                horizon=hor,
                                rotation=rot,
                                openReceptacle=False,
                                pivotId=0,
                                receptacleObjectId='',
                                x=point['x'],
                                y=point['y'],
                                z=point['z'])))

                        if j['openable']:
                            self.step(action=dict(
                                action='OpenObject',
                                forceVisible=True,
                                objectId=j['objectId']),
                                      raise_for_failure=True)
                        for pivot_id in range(j['receptacleCount']):
                            self.step(
                                action=dict(
                                    action='Replace',
                                    forceVisible=True,
                                    receptacleObjectId=j['objectId'],
                                    objectId=visibility_object_id,
                                    pivot=pivot_id), raise_for_failure=True)
                            if self.is_object_visible(visibility_object_id):
                                receptacle_pivot_points.append(dict(
                                    distance=j['distance'],
                                    pivotId=pivot_id,
                                    receptacleObjectId=j['objectId'],
                                    searchNode=dict(
                                        horizon=hor,
                                        rotation=rot,
                                        openReceptacle=j['openable'],
                                        pivotId=pivot_id,
                                        receptacleObjectId=j['objectId'],
                                        x=point['x'],
                                        y=point['y'],
                                        z=point['z'])))

                        if j['openable']:
                            self.step(action=dict(
                                action='CloseObject',
                                forceVisible=True,
                                objectId=j['objectId']),
                                      raise_for_failure=True)

        return receptacle_pivot_points, receptacle_points

    def find_visible_objects(self):

        seen_target_objects = defaultdict(list)

        for point in self.grid_points:
            self.step(dict(
                action='Teleport',
                x=point['x'],
                y=point['y'],
                z=point['z']), raise_for_failure=True)

            for rot, hor in product(self.rotations, self.horizons):
                event = self.step(dict(
                    action='RotateLook',
                    rotation=rot,
                    horizon=hor), raise_for_failure=True)

                object_receptacle = dict()
                for obj in event.metadata['objects']:
                    if obj['receptacle']:
                        for pso in obj['pivotSimObjs']:
                            object_receptacle[pso['objectId']] = obj

                for obj in filter(
                        lambda x: x['visible'] and x['pickupable'],
                        event.metadata['objects']):

                    #if obj['objectId'] in object_receptacle and\
                    #        object_receptacle[obj['objectId']]['openable'] and not \
                    #        object_receptacle[obj['objectId']]['isopen']:
                    #    continue

                    seen_target_objects[obj['objectId']].append(dict(
                        distance=obj['distance'],
                        agent=event.metadata['agent']))

        return seen_target_objects

    def initialize_scene(self):
        self.target_objects = []
        self.object_receptacle = defaultdict(
            lambda: dict(objectId='StartupPosition', pivotSimObjs=[]))

        self.open_receptacles = []
        open_pickupable = {}
        pickupable = {}
        is_open = {}

        for obj in filter(lambda x: x['receptacle'], self.last_event.metadata['objects']):
            for oid in obj['receptacleObjectIds']:
                self.object_receptacle[oid] = obj

            is_open[obj['objectId']] = (obj['openable'] and obj['isOpen'])

        for obj in filter(lambda x: x['receptacle'], self.last_event.metadata['objects']):
            for oid in obj['receptacleObjectIds']:
                if obj['openable'] or (obj['objectId'] in self.object_receptacle and self.object_receptacle[obj['objectId']]['openable']):

                    open_pickupable[oid] = obj['objectId']
                else:
                    pickupable[oid] = obj['objectId']

        if open_pickupable.keys():
            self.target_objects = random.sample(open_pickupable.keys(), k=1)
            shuffled_keys = list(open_pickupable.keys())
            random.shuffle(shuffled_keys)
            for oid in shuffled_keys:
                position_target = self.object_receptacle[self.target_objects[0]]['position']
                position_candidate = self.object_receptacle[oid]['position']
                dist = math.sqrt(
                    (position_target['x'] - position_candidate['x']) ** 2 +
                    (position_target['y'] - position_candidate['y']) ** 2)
                # try to find something that is far to avoid having the doors collide
                if dist > 1.25:
                    self.target_objects.append(oid)
                    break

        for roid in set(map(lambda x: open_pickupable[x], self.target_objects)):
            if roid in is_open:
                continue
            self.open_receptacles.append(roid)
            self.step(dict(
                action='OpenObject',
                objectId=roid,
                forceVisible=True), raise_for_failure=True)

    def queue_step(self):
        search_point = self.queue.popleft()
        event = self.step(dict(
            action='Teleport',
            x=search_point.start_position['x'],
            y=search_point.start_position['y'],
            z=search_point.start_position['z']))

        if event.metadata['lastActionSuccess']:
            move_vec = search_point.move_vector
            move_vec['moveMagnitude'] = self.grid_size
            event = self.step(dict(action='Move', **move_vec))
            if event.metadata['agent']['position']['y'] > 1.3:
                #pprint(search_point.start_position)
                #pprint(search_point.move_vector)
                #pprint(event.metadata['agent']['position'])
                raise Exception("**** got big point ")

            self.enqueue_points(event.metadata['agent']['position'])

            if not any(map(lambda p: distance(p, event.metadata['agent']['position']) < self.distance_threshold, self.grid_points)):
                self.grid_points.append(event.metadata['agent']['position'])

        return event



import re

def scene_names_key_func(scene_name):
    """
    Key function for sorting scenes with the naming convention that was used
    """
    m = re.search('FloorPlan[_]?([a-zA-Z\-]*)([0-9]+)_?([0-9]+)?.*$', scene_name)
    last_val = m.group(3) if m.group(3) is not None else -1
    return m.group(1), int(m.group(2)), int(last_val)


# Copyright Allen Institute for Artificial Intelligence 2017
import requests
from progressbar import ProgressBar, Bar, Percentage, FileTransferSpeed
import hashlib
import logging
import os
import ai2thor.build
from aws_requests_auth.boto_utils import BotoAWSRequestsAuth

logger = logging.getLogger(__name__)

def download(url, build_name, sha256_digest, include_private_scenes=False):

    auth = None
    if include_private_scenes:
        auth = ai2thor.build.boto_auth()

    logger.debug("Downloading file from %s" % url)
    r = requests.get(url, stream=True, auth=auth)
    r.raise_for_status()
    size = int(r.headers['Content-Length'].strip())
    total_bytes = 0

    widgets = [
        build_name, ": ", Bar(marker="|", left="[", right=" "),
        Percentage(), " ", FileTransferSpeed(), "]  of {0}MB".format(str(round(size / 1024 / 1024, 2))[:4])]

    pbar = ProgressBar(widgets=widgets, maxval=size).start()
    m = hashlib.sha256()
    file_data = []
    for buf in r.iter_content(1024):
        if buf:
            file_data.append(buf)
            m.update(buf)
            total_bytes += len(buf)
            pbar.update(total_bytes)
    pbar.finish()
    if m.hexdigest() != sha256_digest:
        raise Exception("Digest mismatch for url %s" % url)

    return b''.join(file_data)
